/*
** Module   : MouseEvt.java
** Abstract : Container for a mouse event, to be processed later by the 4GL logic.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.  ThinClient is explicitly imported.
** 002 CA  20141021 Added API to translate from OS mouse event to legacy 4GL mouse event(s).
** 003 CA  20141029 Switched from event codes to event labels, when raising a portable mouse event
**                  (as the event codes are dependent on the client's compatibility mode (ChUI/GUI,
**                  Windows/Linux).
** 004 EVL 20150401 Making mouse event the extending of the KeyInput.
*/

package com.goldencode.p2j.ui.client.event;

import java.awt.event.*;
import java.util.*;

import com.goldencode.p2j.ui.Keyboard;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;

import static com.goldencode.p2j.ui.Keyboard.*;

/**
 * Container for mouse events.  Instances of this class are posted via 
 * {@link ThinClient#postMouseEvent} to a specialized event queue only if the business logic is
 * in a event processing loop, pause or readkey.
 */
public class MouseEvt
extends KeyInput
{
   /** Mouse click event. */
   public static final int MOUSE_CLICKED = MouseEvent.MOUSE_CLICKED;
   
   /** Mouse press event. */
   public static final int MOUSE_PRESSED = MouseEvent.MOUSE_PRESSED;
   
   /** Mouse release event. */
   public static final int MOUSE_RELEASED = MouseEvent.MOUSE_RELEASED;
   
   /** The captured AWT mouse event. */
   private final MouseEvent event;

   /**
    * Create a new mouse event in the specified window.
    * 
    * @param    event
    *           The AWT mouse event.
    * @param    window
    *           The window where the event originated.
    */
   public MouseEvt(MouseEvent event, TitledWindow<?> window)
   {
      super(window, -1);
      
      this.event = event;
   }
   
   /**
    * Get the AWT mouse event.
    * 
    * @return   See above.
    */
   public MouseEvent getEvent()
   {
      return event;
   }

   /**
    * Get the legacy 4GL events generated by this OS mouse event.  Note that the physical mouse
    * events (i.e. LEFT_MOUSE_*, RIGHT_MOUSE_*) are just aliases for the portable mouse events.
    * <p>
    * The rules to transform a OS mouse event into a portable mouse event are:
    * <ul>
    * 
    * <li>if the {@link MouseEvent#BUTTON1} is pressed (left-mouse-button):
    * <ol>
    * <li>{@link MouseEvent#MOUSE_PRESSED} and {@link MouseEvent#MOUSE_RELEASED} are ignored if
    *     {@link MouseEvent#getClickCount()} is not zero.</li>
    * <li>{@link MouseEvent#MOUSE_PRESSED} is transformed into a 
    *     {@link Keyboard#MOUSE_SELECT_DOWN} or {@link Keyboard#MOUSE_EXTEND_DOWN}, if CTRL
    *     is pressed.</li>
    * <li>{@link MouseEvent#MOUSE_RELEASED} is transformed into a 
    *     {@link Keyboard#MOUSE_SELECT_UP} or {@link Keyboard#MOUSE_EXTEND_UP}, if CTRL
    *     is pressed.</li>
    * <li>{@link MouseEvent#MOUSE_CLICKED} is transformed into a 
    *     {@link Keyboard#MOUSE_SELECT_DOWN} followed by a {@link Keyboard#MOUSE_SELECT_UP},
    *     {@link Keyboard#MOUSE_EXTEND_DOWN} followed by a {@link Keyboard#MOUSE_EXTEND_UP},
    *     if CTRL is pressed.</li>
    * <li>if a double-click is generated, then another {@link Keyboard#MOUSE_SELECT_UP} or
    *     {@link Keyboard#MOUSE_EXTEND_UP} are generated, this time having as function
    *     {@link Keyboard#MOUSE_SELECT_DBLCLICK} or {@link Keyboard#MOUSE_EXTEND_DBLCLICK}.
    * </ol>
    * </li>
    * 
    * <li>if the {@link MouseEvent#BUTTON3} is pressed (right-mouse-button):
    * <ol>
    * <li>{@link MouseEvent#MOUSE_PRESSED} and {@link MouseEvent#MOUSE_RELEASED} are ignored if
    *     {@link MouseEvent#getClickCount()} is not zero.</li>
    * <li>{@link MouseEvent#MOUSE_PRESSED} is transformed into a 
    *     {@link Keyboard#MOUSE_MENU_DOWN}.</li>
    * <li>{@link MouseEvent#MOUSE_RELEASED} is transformed into a 
    *     {@link Keyboard#MOUSE_MENU_UP}.</li>
    * <li>{@link MouseEvent#MOUSE_CLICKED} is transformed into a 
    *     {@link Keyboard#MOUSE_MENU_DOWN} followed by a {@link Keyboard#MOUSE_MENU_UP}.
    *     </li>
    * <li>if a double-click is generated, then another {@link Keyboard#MOUSE_MENU_UP} is
    *     generated, this time having as function {@link Keyboard#MOUSE_MENU_DBLCLICK}.
    * </ol>
    * </li>
    * 
    * <li>middle-button events ({@link MouseEvent#BUTTON2}) are not supported by 4GL, even if
    *     triggers can be defined for these events.</li>
    * 
    * <li>KEYCODE(MOUSE_MOVE_*) for portable events reports the same codes as for the 
    *     MOUSE_SELECT_* events.  These might not be real events but just actions triggered for
    *     the MOUSE_SELECT_* events (in certain circumstances), but this can't be duplicated yet.
    *     TODO: find the correct behavior for these events.</li> 
    * </ul>
    * @param    source
    *           The source widget.
    * 
    * @return   A list of {@link PortableMouseEvent events}.
    */
   public List<PortableMouseEvent> legacyEvents(Widget<?> source)
   {
      List<PortableMouseEvent> evts = new ArrayList<>();
      int button = event.getButton();
      int clicks = event.getClickCount();
      boolean ctrl = event.isControlDown();

      switch (event.getID())
      {
         case MouseEvent.MOUSE_CLICKED:
            switch (button)
            {
               case MouseEvent.BUTTON1:
                  evts.add(new PortableMouseEvent(source, 
                                                  ctrl ? MOUSE_EXTEND_DOWN 
                                                       : MOUSE_SELECT_DOWN, 
                                                  -1)); 
                  evts.add(new PortableMouseEvent(source, 
                                                  ctrl ? MOUSE_EXTEND_UP 
                                                       : MOUSE_SELECT_UP, 
                                                  ctrl ? MOUSE_EXTEND_CLICK 
                                                       : MOUSE_SELECT_CLICK));
                  
                  if (clicks == 2)
                  {
                     evts.add(new PortableMouseEvent(source,
                                                     ctrl ? MOUSE_EXTEND_UP   
                                                          : MOUSE_SELECT_UP,
                                                     ctrl ? MOUSE_EXTEND_DBLCLICK 
                                                          : MOUSE_SELECT_DBLCLICK));
                  }
                  break;

               case MouseEvent.BUTTON3:
                  evts.add(new PortableMouseEvent(source, MOUSE_MENU_DOWN)); 
                  evts.add(new PortableMouseEvent(source, MOUSE_MENU_UP, MOUSE_MENU_CLICK));
                  
                  if (clicks == 2)
                  {
                     evts.add(new PortableMouseEvent(source,
                                                     MOUSE_MENU_UP,
                                                     MOUSE_MENU_DBLCLICK));
                  }
                  break;
            }
            break;

         case MouseEvent.MOUSE_PRESSED:
            if (clicks != 0)
            {
               // wait for the OS mouse-click event
               break;
            }

            switch (button)
            {
               case MouseEvent.BUTTON1:
                  evts.add(new PortableMouseEvent(source, 
                                                  ctrl ? MOUSE_EXTEND_DOWN 
                                                       : MOUSE_SELECT_DOWN, 
                                                 -1)); 
                  break;
               case MouseEvent.BUTTON3:
                  evts.add(new PortableMouseEvent(source, MOUSE_MENU_DOWN)); 
                  break;
            }
            break;
         
         case MouseEvent.MOUSE_RELEASED:
            if (clicks != 0)
            {
               // wait for the OS mouse-click event
               break;
            }

            switch (button)
            {
               case MouseEvent.BUTTON1:
                  evts.add(new PortableMouseEvent(source, 
                                                  ctrl ? MOUSE_EXTEND_UP 
                                                       : MOUSE_SELECT_UP)); 
                  break;
               case MouseEvent.BUTTON3:
                  evts.add(new PortableMouseEvent(source, MOUSE_MENU_UP)); 
                  break;
            }
            break;
      }
      
      return evts;
   }
}
